/*-----copyright-----
 *     amy-studio - make development easier 
 *     web        : http://www.amy-studio.com
 *     taobao     : http://amy-studio.taobao.com
 *     qq(e_mail) : amystudio@qq.com
 *-----file info-----
 *     name       : lcd1602_clock.v
 *     author     : Yuphone Chang
 *     description: a simple clock, display on LCD1602
 *----- version -----
 *     created    : 7.20, 2010
 *     version    : 1.0
 *     description: original version
 *
 *     modified   : 
 *     version    : 
 *     description: 
 *-------------------
 */

module lcd1602_clock(
  input        CLOCK_50,                // global clock 
  input        Q_KEY,                   // asynchronous reset  
  // LCD1602 Interface
  output [7:0] LCD1602_DATA,            // data bus               
  output       LCD1602_E,               // chip enable 
  output       LCD1602_RS,              // data/instruction code 
  output       LCD1602_RW               // read/write 
);

//++++++++++++++++++++++++++++++++++++++
// accessing 1Hz clock begins 
//++++++++++++++++++++++++++++++++++++++
wire clk_1_sec;

div_50M div_50M_u0(
  .clk              (CLOCK_50),
  .rst_n            (Q_KEY),
  .o_clk            (clk_1_sec)
);
//--------------------------------------
// accessing 1Hz clock ends 
//--------------------------------------


//++++++++++++++++++++++++++++++++++++++
// the part of a simple clock begins
//++++++++++++++++++++++++++++++++++++++
// BCD coding, easy to display 
reg [7:0] second_bcd;                   // second 
reg [7:0] minute_bcd;                   // minute 
reg [7:0] hour_bcd;                     // hour 

always @ (posedge clk_1_sec, negedge Q_KEY)
  if (!Q_KEY)
  begin
    second_bcd <= 0;
    minute_bcd <= 0;
    hour_bcd   <= 0;
  end
  else
  begin
    // sencond 00 ~ 59
    if (second_bcd[3:0] == 9)
    begin
      second_bcd[3:0] <= 0;
      second_bcd[7:4] <= second_bcd[7:4] + 1'b1;
      if (second_bcd[7:4] == 5)
        second_bcd[7:4] <= 0;
    end
    else
      second_bcd[3:0] <= second_bcd[3:0] + 1'b1;
    
    // minute 00 ~ 59
    if (second_bcd == 8'h59)
    begin
      if (minute_bcd[3:0] == 9)
      begin
        minute_bcd[3:0] <= 0;
        minute_bcd[7:4] <= minute_bcd[7:4] + 1'b1;
        if (minute_bcd[7:4] == 5)
          minute_bcd[7:4] <= 0;
      end
      else
        minute_bcd[3:0] <= minute_bcd[3:0] + 1'b1; 
    end
    
    // hour 00 ~ 23
    if (minute_bcd == 8'h59 && second_bcd == 8'h59)
    begin
      if (hour_bcd[3:0] == 9 || (hour_bcd[7:4] == 2 && hour_bcd[3:0] == 3))
      begin
        hour_bcd[3:0] <= 0;
        hour_bcd[7:4] <= hour_bcd[7:4] + 1'b1;
        if (hour_bcd[7:4] == 2)
          hour_bcd[7:4] <= 0;
      end
      else
        hour_bcd[3:0] <= hour_bcd[3:0] + 1'b1;
    end
  end

//--------------------------------------
// the part of a simple clock ends 
//--------------------------------------


//++++++++++++++++++++++++++++++++++++++
// displaying the simple clock begins
//++++++++++++++++++++++++++++++++++++++
// 0 ~ (8*16-1) = 128 bits binary code
reg [127:0] row1_val;                   // data for the 1st row
reg [127:0] row2_val;                   // data for the 2nd row

always @ (posedge CLOCK_50, negedge Q_KEY)
  if (!Q_KEY)
  begin
    // 16bits ascii code  
    //       -> 0123456789ABCDEF <-
    row1_val = "A Clock 00:00:00";
    row2_val = "Amy-studio Pub  ";
  end
  else
  // convert BCD code into ASCII code 
  // [+48][refer to ASCII code table]
  begin
    row1_val[7:0]   <= 48 + second_bcd[3:0];
    row1_val[15:8]  <= 48 + second_bcd[7:4];
    // row1_val[23:16] // :    
    row1_val[31:24] <= 48 + minute_bcd[3:0];
    row1_val[39:32] <= 48 + minute_bcd[7:4];
    // row1_val[47:40] // :
    row1_val[55:48] <= 48 + hour_bcd[3:0];
    row1_val[63:56] <= 48 + hour_bcd[7:4];
  end

// Instantiation of LCD1602 display drive module
lcd1602_drive lcd_1602_u0(
  .clk              (CLOCK_50),
  .rst_n            (Q_KEY),
  // LCD1602 Input Value
  .row1_val         (row1_val),
  .row2_val         (row2_val),
  // LCD1602 Interface
  .lcd_data         (LCD1602_DATA),
  .lcd_e            (LCD1602_E),
  .lcd_rs           (LCD1602_RS),
  .lcd_rw           (LCD1602_RW)
);
//--------------------------------------
// displaying the simple clock ends 
//--------------------------------------

endmodule
